#include "owl.h"
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#define INCR 5000

void owl_editwin_init(owl_editwin *e, WINDOW *win, int winlines, int wincols, int style) {
  e->buff=malloc(INCR);
  e->buff[0]='\0';
  e->bufflen=0;
  e->allocated=INCR;
  e->buffx=0;
  e->buffy=0;
  e->topline=0;
  e->winlines=winlines;
  e->wincols=wincols;
  e->curswin=win;
  e->style=style;
  if ((style!=OWL_EDITWIN_STYLE_MULTILINE) &&
      (style!=OWL_EDITWIN_STYLE_ONELINE)) {
    e->style=OWL_EDITWIN_STYLE_MULTILINE;
  }
  e->lock=0;
  if (win) werase(win);
}

void owl_editwin_set_locktext(owl_editwin *e, char *text) {
  /* set text to be 'locked in' at the beginning of the buffer, any
     previous text (including locked text) will be overwritten */
  
  int x, y;

  x=e->buffx;
  y=e->buffy;
  e->buffx=0;
  e->buffy=0;
  owl_editwin_overwrite_string(e, text);
  e->lock=strlen(text);
  /* if (text[e->lock-1]=='\n') e->lock--; */
  e->buffx=x;
  e->buffy=y;
  owl_editwin_adjust_for_locktext(e);
  owl_editwin_redisplay(e);
}

void owl_editwin_set_curswin(owl_editwin *e, WINDOW *w, int winlines, int wincols) {
  e->curswin=w;
  e->winlines=winlines;
  e->wincols=wincols;
}

int owl_editwin_get_style(owl_editwin *e) {
  return(e->style);
}

void owl_editwin_new_style(owl_editwin *e, int newstyle) {
  char *ptr;
  
  if (e->style==newstyle) return;

  if (newstyle==OWL_EDITWIN_STYLE_MULTILINE) {
    e->style=newstyle;
  } else if (newstyle==OWL_EDITWIN_STYLE_ONELINE) {
    e->style=newstyle;

    /* nuke everything after the first line */
    if (e->bufflen > 0) {
      ptr=strchr(e->buff, '\n')-1;
      if (ptr) {
	e->bufflen=ptr - e->buff;
	e->buff[e->bufflen]='\0';
	e->buffx=0;
	e->buffy=0;
      }
    }
  }
  
}

void owl_editwin_fullclear(owl_editwin *e) {
  /* completly reinitialize the buffer */
  free(e->buff);
  owl_editwin_init(e, e->curswin, e->winlines, e->wincols, e->style);
}

void owl_editwin_clear(owl_editwin *e) {
  /* clear all text except for locktext and put the cursor at the beginning */
  int lock;
  char *locktext=NULL;
  
  lock=0;
  if (e->lock > 0) {
    lock=1;

    locktext=malloc(e->lock+20);
    strncpy(locktext, e->buff, e->lock);
    locktext[e->lock]='\0';
  }

  free(e->buff);
  owl_editwin_init(e, e->curswin, e->winlines, e->wincols, e->style);

  if (lock > 0) {
    owl_editwin_set_locktext(e, locktext);
  }

  if (locktext) free(locktext);
  owl_editwin_adjust_for_locktext(e);
}


void _owl_editwin_addspace(owl_editwin *e) {
  /* malloc more space for the buffer */
  e->buff=realloc(e->buff, e->allocated+INCR);
  if (!e->buff) {
    /* error */
    return;
  }
  e->allocated+=INCR;
}

void owl_editwin_redisplay(owl_editwin *e) {
  /* regenerate the text on the curses window */
  char *ptr1, *ptr2, *ptr3, *buff;
  int i;

  werase(e->curswin);
  wmove(e->curswin, 0, 0);

  /* start at topline */
  ptr1=e->buff;
  for (i=0; i<e->topline; i++) {
    ptr2=strchr(ptr1, '\n');
    if (!ptr2) {
      /* we're already on the last line */
      break;
    }
    ptr1=ptr2+1;
  }
  /* ptr1 now stores the starting point */

  /* find the ending point and store it in ptr3 */
  /* just ignore columns for now */
  ptr2=ptr1;
  ptr3=ptr1;
  for (i=0; i<e->winlines; i++) {
    ptr3=strchr(ptr2, '\n');
    if (!ptr3) {
      /* we've hit the last line */
      /* print everything to the end */
      ptr3=e->buff+e->bufflen-1;
      ptr3--;
      break;
    }
    ptr2=ptr3+1;
  }
  ptr3+=2;

  buff=malloc(ptr3-ptr1+50);
  strncpy(buff, ptr1, ptr3-ptr1);
  buff[ptr3-ptr1]='\0';
  waddstr(e->curswin, buff);
  wmove(e->curswin, e->buffy-e->topline, e->buffx);
  wrefresh(e->curswin);
  free(buff);
}


void owl_editwin_insert_char(owl_editwin *e, char c) {
  /* insert a character at the current point (shift later
   * characters over */
  
  int z, i;

  /* \r is \n */
  if (c=='\r') {
    c='\n';
  }

  if (c=='\n' && e->style==OWL_EDITWIN_STYLE_ONELINE) {
    /* perhaps later this will change some state that allows the string
       to be read */
    return;
  }

  /* make sure there is enough memory for the new text */
  if ((e->bufflen+1) > (e->allocated-5)) {
    _owl_editwin_addspace(e);
  }

  /* get the insertion point */
  z=_owl_editwin_get_index_from_xy(e);

  /* shift */
  for (i=e->bufflen; i>z; i--) {
    e->buff[i]=e->buff[i-1];
  }

  e->buff[z]=c;

  /* housekeeping */
  e->bufflen++;
  e->buff[e->bufflen]='\0';

  /* advance the cursor */
  if (c=='\n') {
    e->buffx=0;
    e->buffy++;
  } else {
    e->buffx++;
  }

}

void owl_editwin_overwrite_char(owl_editwin *e, char c) {
  /* overwrite the character at the current point with 'c' */
  int z;
  
  /* \r is \n */
  if (c=='\r') {
    c='\n';
  }

  if (c=='\n' && e->style==OWL_EDITWIN_STYLE_ONELINE) {
    /* perhaps later this will change some state that allows the string
       to be read */
    return;
  }

  z=_owl_editwin_get_index_from_xy(e);

  /* only if we are at the end of the buffer do we create new space */
  if (z==e->bufflen) {
    if ((e->bufflen+1) > (e->allocated-5)) {
      _owl_editwin_addspace(e);
    }
  }
  
  e->buff[z]=c;

  /* housekeeping if we are at the end of the buffer */
  if (z==e->bufflen) {
    e->bufflen++;
    e->buff[e->bufflen]='\0';
  }

  /* advance the cursor */
  if (c=='\n') {
    e->buffx=0;
    e->buffy++;
  } else {
    e->buffx++;
  }

}

void owl_editwin_delete_char(owl_editwin *e) {
  /* delete the character at the current point, following chars
   * shift left.
   */ 
  int z, i;

  if (e->bufflen==0) return;
  
  /* get the deletion point */
  z=_owl_editwin_get_index_from_xy(e);

  for (i=z; i<e->bufflen; i++) {
    e->buff[i]=e->buff[i+1];
  }
  e->bufflen--;
  e->buff[e->bufflen]='\0';
}

void owl_editwin_insert_string(owl_editwin *e, char *string) {
  /* insert 'string' at the current point, later text is shifted
   * right
   */
  int i, j;

  j=strlen(string);
  for (i=0; i<j; i++) {
    owl_editwin_insert_char(e, string[i]);
  }
}

void owl_editwin_overwrite_string(owl_editwin *e, char *string) {
  int i, j;
  /* write 'string' at the current point, overwriting text that is
   * already there
   */

  j=strlen(string);
  for (i=0; i<j; i++) {
    owl_editwin_overwrite_char(e, string[i]);
  }
}

int _owl_editwin_get_index_from_xy(owl_editwin *e) {
  /* get the index into e->buff for the current cursor
   * position.
   */
  int i;
  char *ptr1, *ptr2;

  if (e->bufflen==0) return(0);
  
  /* first go to the yth line */
  ptr1=e->buff;
  for (i=0; i<e->buffy; i++) {
    ptr2=strchr(ptr1, '\n');
    if (!ptr2) {
      /* we're already on the last line */
      break;
    }
    ptr1=ptr2+1;
  }

  /* now go to the xth character */
  ptr2=strchr(ptr1, '\n');
  if (!ptr2) {
    ptr2=e->buff+e->bufflen;
  }

  if ((ptr2-ptr1) < e->buffx) {
    ptr1=ptr2-1;
  } else {
    ptr1+=e->buffx;
  }

  /* printf("DEBUG: index is %i\r\n", ptr1-e->buff); */
  return(ptr1-e->buff);
}

void owl_editwin_adjust_for_locktext(owl_editwin *e) {
  /* if we happen to have the cursor over locked text
   * move it to be out of the locktext region */
  while (_owl_editwin_get_index_from_xy(e)<e->lock) {
    owl_editwin_key_right(e);
  }
}

void owl_editwin_backspace(owl_editwin *e) {
  /* delete the char before the current one
   * and shift later chars left
   */
  if (_owl_editwin_get_index_from_xy(e) > e->lock) {
    owl_editwin_key_left(e);
    owl_editwin_delete_char(e);
  }
  owl_editwin_adjust_for_locktext(e);
}

void owl_editwin_key_up(owl_editwin *e) {
  if (e->buffy > 0) e->buffy--;
  if (e->buffx >= owl_editwin_get_numchars_on_line(e, e->buffy)) {
    e->buffx=owl_editwin_get_numchars_on_line(e, e->buffy);
  }

  /* do we need to scroll? */
  if (e->buffy-e->topline < 0) {
    e->topline-=e->winlines/2;
  }

  owl_editwin_adjust_for_locktext(e);
}

void owl_editwin_key_down(owl_editwin *e) {
  /* move down if we can */
  if (e->buffy+1 < owl_editwin_get_numlines(e)) e->buffy++;

  /* if we're past the last character move back */
  if (e->buffx >= owl_editwin_get_numchars_on_line(e, e->buffy)) {
    e->buffx=owl_editwin_get_numchars_on_line(e, e->buffy);
  }

  /* do we need to scroll? */
  if (e->buffy-e->topline > e->winlines) {
    e->topline+=e->winlines/2;
  }

  /* adjust for locktext */
  owl_editwin_adjust_for_locktext(e);
}

void owl_editwin_key_left(owl_editwin *e) {
  /* move left if we can, and maybe up a line */
  if (e->buffx>0) {
    e->buffx--;
  } else if (e->buffy>0) {
    e->buffy--;
    e->buffx=owl_editwin_get_numchars_on_line(e, e->buffy);
  }

  /* do we need to scroll up? */
  if (e->buffy-e->topline < 0) {
    e->topline-=e->winlines/2;
  }

  /* make sure to avoid locktext */
  owl_editwin_adjust_for_locktext(e);
}

void owl_editwin_key_right(owl_editwin *e) {
  int i;

  /* move right if we can, and skip down a line if needed */
  i=owl_editwin_get_numchars_on_line(e, e->buffy);
  if (e->buffx < i) {
    e->buffx++;
  } else if (e->buffy+1 <= owl_editwin_get_numlines(e)) {
    if (e->style==OWL_EDITWIN_STYLE_MULTILINE) {
      e->buffx=0;
      e->buffy++;
    }
  }

  /* do we need to scroll down? */
  if (e->buffy-e->topline >= e->winlines) {
    e->topline+=e->winlines/2;
  }
}

void owl_editwin_delete_to_endofline(owl_editwin *e) {
  int i;

  if (owl_editwin_get_numchars_on_line(e, e->buffy)>e->buffx) {
    /* normal line */
    i=_owl_editwin_get_index_from_xy(e);
    while(i < e->bufflen) {
      if (e->buff[i]!='\n') {
	owl_editwin_delete_char(e);
      } else {
	return;
      }
    }
  } else if (e->buffy+1 < owl_editwin_get_numlines(e)) {
    /* line with cursor at the end but not on very last line */
    owl_editwin_key_right(e);
    owl_editwin_backspace(e);
  }
}

void owl_editwin_process_char(owl_editwin *e, int j, int meta) {
  if (j == ERR) return;

  /* cursor movements */
  if (!meta) {
    if ((j==KEY_UP) || (j==16)) {
      if (e->style==OWL_EDITWIN_STYLE_ONELINE) return;
      owl_editwin_key_up(e);
      owl_editwin_redisplay(e);
    } else if ((j==KEY_DOWN) || (j==14)) {
      if (e->style==OWL_EDITWIN_STYLE_ONELINE) return;
      owl_editwin_key_down(e);
      owl_editwin_redisplay(e);
    } else if ((j==KEY_RIGHT) || (j==6)) {
      owl_editwin_key_right(e);
      owl_editwin_redisplay(e);
      return;
    } else if ((j==KEY_LEFT) || (j==2)) {
      owl_editwin_key_left(e);
      owl_editwin_redisplay(e);
      return;
    }
  }

  /* backspace */
  if (!meta) {
    if (j == 8 || j == 127 || j == KEY_BACKSPACE || j== KEY_DC) {
      owl_editwin_backspace(e);
      owl_editwin_redisplay(e);
      return;
    }
  }

  if (!meta) {
    if (j == 21) {
      /* C-u */
      owl_editwin_clear(e);
      owl_editwin_adjust_for_locktext(e);
      owl_editwin_redisplay(e);
    } else if (j == 1) {
      /* C-a */
      e->buffx=0;
      owl_editwin_adjust_for_locktext(e);
      owl_editwin_redisplay(e);
    } else if (j == 5) {
      /* C-e */
      e->buffx=owl_editwin_get_numchars_on_line(e, e->buffy);
    owl_editwin_redisplay(e);
    } else if (j == 11) {
      /* C-k */
      owl_editwin_delete_to_endofline(e);
      owl_editwin_redisplay(e);
    }
  }
  
  /* characteres to ignore */
  if (j>127 || ((j<32) && (j!=10) && (j!=13))) return;
  if (meta) return;

  owl_editwin_insert_char(e, j);

  /* check if we need to scroll down on enter */
  if (j==13 || j==10) {
    if (e->buffy-e->topline >= e->winlines) {
      e->topline+=e->winlines/2;
    }
  }

  owl_editwin_redisplay(e);
}

char *owl_editwin_get_text(owl_editwin *e) {
  return(e->buff+e->lock);
}

int owl_editwin_get_numchars_on_line(owl_editwin *e, int line) {
  int i;
  char *ptr1, *ptr2;

  if (e->bufflen==0) return(0);
  
  /* first go to the yth line */
  ptr1=e->buff;
  for (i=0; i<line; i++) {
    ptr2=strchr(ptr1, '\n');
    if (!ptr2) {
      /* we're already on the last line */
      return(0);
    }
    ptr1=ptr2+1;
  }

  /* now go to the xth character */
  ptr2=strchr(ptr1, '\n');
  if (!ptr2) {
    return(e->buff + e->bufflen - ptr1);
  }
  return(ptr2-ptr1); /* don't count the newline for now */
}

int owl_editwin_get_numlines(owl_editwin *e) {
  return(owl_text_num_lines(e->buff));
}

